/*********************************************************************/
/* join.c                                                            */
/* function neblo_join()                                             */
/* see COPYING in parent directory                                   */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>


/****************************************************************************

"join <requestor_id> <requestor_id_src> <requestor_ipaddr> <requestor_port>
<cmd_handle> <requestor_keymod> <requestor_keyexp>"

connect algorithm, phase 2.
Let T be the foreign node, ID(T) its id., S this peer:
If ID(T) falls into the interval of addresses stored at S, then S welcomes T
in the ring.
Otherwise, S applies the routing rule (neighs, and fingers) to ID(T), so as to
find another peer P who might be a better candidate for welcoming T;
then S forwards the join request to P.
If ID(T) is not "genuine" (it doesn't match its source through RIPEMD16), S
refuses to manage the request.

This req is generally managed in a recursive style (through a chain of
recipients), so a <cmd_handle> is needed for the last recipient to answer
back to the original <requestor>.

*****************************************************************************/


static void *
welcome_foreign
( void *data )
{

  /*
   *  The foreign peer becomes our new successor in the ring.
   * -) pass our current neighbour list to foreign peer;
   * -) adjust our neighbour list to include foreign peer as new successor
   *    (this may imply discarding the rightmost entry of neighbour list,
   *    if full);
   * -) divide my addr interval in two parts and assign second part
   *    to the foreign; second part starts from foreign id. so the two
   *    parts usually have different size;
   * -) get pubkey from the foreign peer, and challenge;
   * -) open a connection towards new successor, get a connection from it
   *    as well.
   */

  char                 *buf;
  ULL                   a,b;
  neblo_addr            foreign_id;
  neblo_addr            foreign_id_src;
  struct in_addr        foreign_addr;
  unsigned long         faddr;
  u_int16_t             foreign_port;
  unsigned int          fport;
  u_int32_t             foreign_join_handle_val;
  char                 *foreign_keymod;
  char                 *foreign_keyexp;
  RSA                  *foreign_pubkey;
  int                   rc = 0;
  csocket_t            *cs;
  unsigned int          i;
  peer_t               *f,*old_succ;
  neblo_addr            new_up_bound,foreign_up_bound;
  ringofptrs_t         *alt_ring;
  peerptr_t            *tail,*ff;
  char                 *neigh_keymod;
  char                 *neigh_keyexp;
  neblo_addr            old_up_bound;


#ifdef DEBUG_J
  S(fprintf(stderr,"welcome_foreign() thread...\n"));
#endif

  buf = (char *)data;

  _do_(_malloc_(foreign_keymod,KEYBUFLEN+SLACK),
       !foreign_keymod, "welcome_foreign(): foreign_keymod malloc");
  _do_(_malloc_(foreign_keyexp,BUFLEN+SLACK),
       !foreign_keyexp, "welcome_foreign(): foreign_keyexp malloc");

  sscanf(buf,"%*s %016llx %016llx %lu %u %u %50000s %100s",
         &a,&b,&faddr,&fport,&foreign_join_handle_val,
         foreign_keymod,foreign_keyexp);

  foreign_id = a;
  foreign_id_src = b;
  foreign_addr.s_addr = faddr;
  foreign_port = (u_int16_t)fport;

  /*  Prepare to open a csock and talk to the newcomer.  */

  _do_(_RSA_new_(foreign_pubkey),!foreign_pubkey,
       "welcome_foreign(): RSA_new of foreign_pubkey");
  _convert_pubkey_(foreign_pubkey,foreign_keymod,foreign_keyexp,rc = -1);
  if (rc < 0) {

#ifdef DEBUG_J
    S(fprintf(stderr,"welcome_foreign(): bad foreign pubkey\n"));
#endif

    goto abort0;
  }

  /*
   * Build a list of ptr to peers currently belonging to neigh list.
   * This will allow us to let the neigh list be concurrently changed
   * while this routine is still running.
   *
   * NOTE: this introduces aliasing on all peers in neigh list.
   *
   * NOTE: the temp copy ensures safety w.r.t. addition of new entries in
   * neigh list concurrently performed by check_routes().  However,
   * due to the aforementioned aliasing, we are still unsafe w.r.t.
   * deletion of entries operated by check_routes() itself
   * (deletion is followed by free() which is harmful in case of aliasing).
   * That's why ("that's why"...) we must raise the "dont_clean" now.
   */

  pthread_mutex_lock(&clean_lock);
  dont_clean++;
  pthread_mutex_unlock(&clean_lock);

  _do_(_calloc_(alt_ring,sizeof(ringofptrs_t)),!alt_ring,
       "welcome_foreign(): alt_ring malloc");

  pthread_mutex_lock(&ring->lock);
  memcpy(&alt_ring->ring,ring,sizeof(ring_t));
  old_succ = ring->neigh_list;
  pthread_mutex_unlock(&ring->lock);

  alt_ring->peerlist = tail = NULL;

  forall_neighbours(f,
    _do_(_calloc_(ff,sizeof(peerptr_t)),!ff,"welcome_foreign(): ff calloc");
    ff->f = f;
    if (f == ring->neigh_list)
      alt_ring->peerlist = tail = ff;
    else {
      tail->next = ff;
      tail = ff;
    }
    ff->next = NULL;
  );

  /*
   * Notify our current neighbours (successor list) about the newcomer.
   * Notifying our current successor is a complicated matter.
   * We don't want to break the successor relation before we are sure
   * about the newcomer has entered.  So, for now, we do not notify our
   * successor; we just make a liveness check.
   * Only the first MAXNEIGH neighbours are considered.
   */

  for (i=MAXNEIGH, ff=alt_ring->peerlist; i && ff!=NULL; i--, ff=ff->next) {

    f = ff->f;

    if (!f->self) {
      if (is_successor(f)) {

#ifdef DEBUG_J
        S(fprintf(stderr,
                  "liveness checking successor (%u.%u.%u.%u:%u, status %u)\n",
                  ((unsigned char *)&f->addr.s_addr)[0],
                  ((unsigned char *)&f->addr.s_addr)[1],
                  ((unsigned char *)&f->addr.s_addr)[2],
                  ((unsigned char *)&f->addr.s_addr)[3],
                  f->port,f->status));
#endif

        rc = tell_peer(f,1,"ping",strlen("ping")+1,0,0 /* dummy th_id */);
        bzero(buf,5);
        rc = listen_peer(f,buf,strlen("ping")+1,TIMEOUT,0);
        rc = strncmp(buf,"ping",strlen("ping"));
        rc = -(rc != 0);  /*  -1 iff strings differ from each other  */
      } else {

#ifdef DEBUG_J
        S(fprintf(stderr,"notifying to %u.%u.%u.%u:%u, status %u\n",
                  ((unsigned char *)&f->addr.s_addr)[0],
                  ((unsigned char *)&f->addr.s_addr)[1],
                  ((unsigned char *)&f->addr.s_addr)[2],
                  ((unsigned char *)&f->addr.s_addr)[3],
                  f->port,f->status));
#endif

        rc = announce_friend(f,foreign_addr,foreign_port,
                             foreign_keymod,foreign_keyexp,
                             0 /* dummy th_id */);
      }
      if (rc < 0) {

#ifdef DEBUG_J
        S(fprintf(stderr,"(%u.%u.%u.%u:%u has troubles...)\n",
                  ((unsigned char *)&f->addr.s_addr)[0],
                  ((unsigned char *)&f->addr.s_addr)[1],
                  ((unsigned char *)&f->addr.s_addr)[2],
                  ((unsigned char *)&f->addr.s_addr)[3],
                  f->port));
#endif

        /*
         *  If the faulting peer is our current successor we cannot
         *  welcome any other peer until a new successor is found.
         */
        if (ff == alt_ring->peerlist)
          break;
        else
          rc = 0;
      }
    } else {
      /*
       *  This covers the startup case, when this peer is the generator of
       *  the ring.
       *  When the peer first generated the ring, it had to include itself
       *  in its own neighbour list, to "keep the ring closed".
       *  As a consequence, each new successor get this peer in its
       *  neighbour list, so it will eventually open another conn. to us.
       */
      if (!empty_neighbour_list())
        insert_expd_reqst(foreign_addr,foreign_port,
                          0 /* new peer is successor, not predecessor */,
                          foreign_pubkey, 0 /* dummy th_id */);
      else
        insert_expd_reqst(foreign_addr,foreign_port,
                          1 /*
                             * empty ring: new peer is successor and
                             * predecessor at the same time
                             */,
                          foreign_pubkey, 0 /* dummy th_id */);
    }

  }

  if (rc < 0) {
    /*
     *  Our current successor has troubles.  We cannot
     *  welcome any peer until a new successor is found.
     */

#ifdef DEBUG_J
    S(fprintf(stderr,"Troubles with current successor (out).  Abort...\n"));
#endif

    goto abort;
  }

  /*
   * We can only accept a newcomer if the bidirectional connection with
   * our current successor is established.  The outgoing connection has
   * been probed, now check if the incoming one exists.
   * Exception: when this peer is alone in the ring (no successor)...
   */
  rc = -1;
  forall_friends(f,if (f->from_successor) {rc = 0; break;});
  if (rc < 0 && ring->friend_list) {

#ifdef DEBUG_J
    S(fprintf(stderr,"Troubles with current successor (in).  Abort...\n"));
#endif

    goto abort;
  }

  /*
   * The relation with our current succ is sane, so we can welcome
   * the foreign peer
   */

  _openout_(foreign_addr,foreign_port,&cs,foreign_pubkey,&rc);
  if (rc < 0) {

#ifdef DEBUG_J
    S(fprintf(stderr,"welcome_foreign(): refused conn from foreign\n"));
#endif

    goto abort;
  }

  sprintf(buf,"welcome %u %u %s %s",
          foreign_join_handle_val,PORT,pubkeymod,pubkeyexp);
  _write_(cs,buf,strlen(buf)+1,rc,goto abort);

#ifdef DEBUG_J
  S(fprintf(stderr,"<< %s\n",buf));
#endif

  /*  Challenge foreign on his supposed pubkey. */
  rc = _put_challenge_(cs,buf,foreign_pubkey);
  if (rc < 0) {

#ifdef DEBUG_J
  S(fprintf(stderr,"challenge failed or interrupted\n"));
#endif

    goto abort;
  }

#ifdef DEBUG_J
  S(fprintf(stderr,"challenge ok\n"));
#endif

  /*
   * the addr interval assigned to this peer
   * splits in two parts: first part remains allocated
   * here, whereas second part becomes allocated to the foreign,
   * because it has greater id. in the 64-bit ring ordering
   */

  foreign_up_bound = ring->up_bound;
  new_up_bound = foreign_id - 1;

  sprintf(buf,"intvl %016llx",(ULL)foreign_up_bound);
  _write_(cs,buf,strlen(buf)+1,rc,goto abort);

#ifdef DEBUG_J
  S(fprintf(stderr,"<< %s\n",buf));
#endif

  /*  Send our current neighbour list to the newcomer.  */

  for (i=MAXNEIGH, ff=alt_ring->peerlist; i && ff!=NULL; i--, ff=ff->next) {

    f = ff->f;

#ifdef DEBUG_J
    S(fprintf(stderr,"introducing %u.%u.%u.%u:%u, status %u\n",
              ((unsigned char *)&f->addr.s_addr)[0],
              ((unsigned char *)&f->addr.s_addr)[1],
              ((unsigned char *)&f->addr.s_addr)[2],
              ((unsigned char *)&f->addr.s_addr)[3],
              f->port,f->status));
#endif

    _do_(_BN_bn2hex_(neigh_keymod,f->pubkey->n),!neigh_keymod,
         "welcome_foreign(): neigh_keymod alloc");
    _do_(_BN_bn2hex_(neigh_keyexp,f->pubkey->e),!neigh_keyexp,
         "welcome_foreign(): neigh_keyexp alloc");
    sprintf(buf,"%lu %u %s %s",
            (unsigned long)f->addr.s_addr,f->port,
            neigh_keymod,neigh_keyexp);
    _OPENSSL_free_(neigh_keymod); _OPENSSL_free_(neigh_keyexp);

    _write_(cs,buf,strlen(buf)+1,rc,break);

#ifdef DEBUG_J
    S(fprintf(stderr,"<< \"neighbour %u.%u.%u.%u:%u\"\n",
              ((unsigned char *)&f->addr.s_addr)[0],
              ((unsigned char *)&f->addr.s_addr)[1],
              ((unsigned char *)&f->addr.s_addr)[2],
              ((unsigned char *)&f->addr.s_addr)[3],
              f->port));
#endif

    bzero(buf,4);
    _read_(cs,buf,BUFLEN,rc,TIMEOUT,break);
    if (strncmp(buf,"ok",strlen("ok"))) {

#ifdef DEBUG_J
      S(fprintf(stderr,"error while welcoming new peer\n"));
#endif

      rc = -1;
      break;
    } else {

#ifdef DEBUG_J
      S(fprintf(stderr,">> %s\n",buf));
#endif

    }

  }

  if (rc < 0) {

#ifdef DEBUG_J
    S(fprintf(stderr,"Troubles with newcomer.  Abort...\n"));
#endif

    goto abort;
  }

  _write_(cs,"end",strlen("end")+1,rc,goto abort);

#ifdef DEBUG_J
  S(fprintf(stderr,"<< end\n"));
#endif

  /*
   *  Foreign node is now about to be inserted into the neighbour list,
   *  as the new successor.
   */

  _do_(_calloc_(f,sizeof(peer_t)),!f,"welcome_foreign(): f calloc");
  _pthmutexinit_(&f->lock);
  f->addr.s_addr = foreign_addr.s_addr;
  f->port = foreign_port;
  f->pubkey = foreign_pubkey;
  /*  All other fields are zero/NULL, by calloc  */

  /*
   *  Establish a bidirectional connection with the new successor
   *  for routing (become a "friend" of foreign).
   */

  /*  Wait for the new successor to be ready...  */
  bzero(buf,4);
  _read_(cs,buf,BUFLEN,rc,TIMEOUT,/* nothing */);
  if (strncmp(buf,"ok",strlen("ok"))) {

#ifdef DEBUG_J
    S(fprintf(stderr,"error while waiting new peer to be ready\n"));
#endif

    goto abort1;
  } else {

#ifdef DEBUG_J
    S(fprintf(stderr,">> %s\n",buf));
#endif

  }

  /* expect an incoming conn. from new successor */
  rc = insert_expd_reqst(f->addr,f->port,2 /* from successor */,f->pubkey,
                         0 /* dummy th_id */);
  if (rc < 0) goto abort1;

  f->status = TOCALL;
  for (i=5; i; i--) {
    rc = call_peer(f,1 /* from predecessor */,0,0 /* dummy th_id */);
    if (rc >= 0)
      /* success. status is set to OK */
      break;
    if (rc == -2) {
      /* new successor has died meanwhile. abort... */

#ifdef DEBUG_J
      S(fprintf(stderr,"new successor dead, abort...\n"));
#endif

      goto abort1;
    }
    /* new successor not responding. wait a moment, then try again */
    sleep(1);
  }
  if (!i) {
    /* new successor not responding for too long.  assume it is dead */

#ifdef DEBUG_J
      S(fprintf(stderr,"new successor not responding, abort...\n"));
#endif

      goto abort1;
  }

  /*
   *  Tell successor that we have connected to it and are ready to accept
   *  its own (backward) connection.
   */
  _write_(cs,"ok",strlen("ok")+1,rc,goto abort1);

#ifdef DEBUG_J
  S(fprintf(stderr,"<< ok\n"));
#endif

  /*
   *  The incoming conn from successor will be managed by a separate
   *  thread.  Successor tells us "ok" once it has connected.
   */
  bzero(buf,4);
  _read_(cs,buf,BUFLEN,rc,TIMEOUT,/* */);
  if (strncmp(buf,"ok",strlen("ok"))) goto abort1;

#ifdef DEBUG_J
  S(fprintf(stderr,">> %s\n",buf));
#endif

  /*  Two-way connection with successor has been established.  */

  /*
   *  Insert new successor at head of our neighbour list, and
   *  update the local addr interval.
   *  Inhibit msgs addressed to the local peer until done,
   *  because the local addr intvl is about to change, and we don't
   *  want to take remote data to be then moved to the new peer.
   */
  pthread_mutex_lock(&ring->msg_lock);

  pthread_mutex_lock(&ring->lock);

  /*  Update addr interval  */
  old_up_bound = ring->up_bound;
  ring->up_bound = new_up_bound;

  /*  Insert at head  */
  f->next = ring->neigh_list;
  ring->neigh_list = f;

  pthread_mutex_unlock(&ring->lock);

  /*
   * Ask the upper layer (application) to migrate the data
   * mapped to the addr interval now assigned to the new peer
   * (primary as well as backup copies).
   */

  rc = republish_addrs(foreign_id,foreign_up_bound,foreign_id,NEBLO_MAXCP);
  if (rc < 0) {

#ifdef DEBUG_J
    S(fprintf(stderr,"error while republishing, abort join...\n"));
#endif

    goto abort2;
  }

  /*
   * So far so good.
   * Now we must allow our new successor to connect to our previous
   * successor, so as the ring shape is restored.  We get a final OK from
   * the new successor when done.
   */

  if (!old_succ->self) {

#ifdef DEBUG_J
    S(fprintf(stderr,"notifying to old successor (%u.%u.%u.%u:%u, status %u)\n",
              ((unsigned char *)&old_succ->addr.s_addr)[0],
              ((unsigned char *)&old_succ->addr.s_addr)[1],
              ((unsigned char *)&old_succ->addr.s_addr)[2],
              ((unsigned char *)&old_succ->addr.s_addr)[3],
              old_succ->port,old_succ->status));
#endif

    rc = announce_predecessor(old_succ,foreign_addr,foreign_port,
                              foreign_keymod,foreign_keyexp,
                              0 /* dummy th_id */);
    if (rc < 0) goto abort2;
  }

  _write_(cs,"ok",strlen("ok")+1,rc,/* */);
  if (rc < 0) goto abort2;

#ifdef DEBUG_J
  S(fprintf(stderr,"<< ok\n"));
#endif

  bzero(buf,4);
  _read_(cs,buf,BUFLEN,rc,TIMEOUT,/* */);
  if (rc < 0) goto abort2;

#ifdef DEBUG_J
  S(fprintf(stderr,">> %s\n",buf));
#endif

  if (strncmp(buf,"ok",strlen("ok"))) goto abort2;

  /*
   * We just got the final OK from newcomer, meaning the ring shape has been
   * restored (in the forward direction, not in the backward yet, but it is
   * enough for now).  We are almost done...
   */

  /* resume accepting msgs */
  pthread_mutex_unlock(&ring->msg_lock);

  /*
   * now we can ask the upper layer (application) to remove the data
   * corresponding to the addr interval now migrated to the new peer
   */
  rc = clean_addrs(foreign_id,foreign_up_bound,NEBLO_MAXCP);
  if (rc < 0)
    S(fprintf(stderr,"error from clean handler!\n"));

  /*  Foreign node entered ring successfully :-)  */

#ifdef DEBUG_J
  S(
  fprintf(stderr,"welcomed a new peer:\n");
  fprintf(stderr,"new addr interval: %016llx -- %016llx\n",
          ring->my_id,ring->up_bound);
  fprintf(stderr,"new neighbour list:\n");
  forall_neighbours(f,
    if (f->status & (OK | TOCALL | PROBABLYDEAD))
      fprintf(stderr,"\t%u.%u.%u.%u:%u (%u)\n",
              ((unsigned char *)&f->addr.s_addr)[0],
              ((unsigned char *)&f->addr.s_addr)[1],
              ((unsigned char *)&f->addr.s_addr)[2],
              ((unsigned char *)&f->addr.s_addr)[3],
              f->port,f->status);
  );
  );
#endif

exit:

  _close_(&cs);

  /* dispose the tmp copy of routing tables */
  for (ff=alt_ring->peerlist; ff!=NULL;) {
    tail = ff->next;
    _free_(ff);
    ff = tail;
  }
  _free_(alt_ring);

  pthread_mutex_lock(&clean_lock);
  dont_clean--;
  pthread_mutex_unlock(&clean_lock);

  /* do not free foreign_pubkey, as it is referenced by new peer desc */

  _free_(foreign_keymod); _free_(foreign_keyexp);

  _free_(buf);
  pthread_mutex_unlock(&ring->join_lock);
  pthread_exit(NULL);

abort2:

  ring->up_bound = old_up_bound;
  ring->neigh_list = f->next;
  pthread_mutex_unlock(&ring->msg_lock);

abort1:

  clean_peer_desc(f);
  /*
   * The above also free() the foreign_pubkey, referenced by f, so we must
   * skip the RSA_free() below.
   */
  _free_(f);
  goto exit;

abort:

  _close_(&cs);

  /* dispose the tmp copy of routing tables */
  for (ff=alt_ring->peerlist; ff!=NULL;) {
    tail = ff->next;
    _free_(ff);
    ff = tail;
  }
  _free_(alt_ring);

  pthread_mutex_lock(&clean_lock);
  dont_clean--;
  pthread_mutex_unlock(&clean_lock);

abort0:

  _RSA_free_(foreign_pubkey);
  _free_(foreign_keymod); _free_(foreign_keyexp);

  _free_(buf);
  pthread_mutex_unlock(&ring->join_lock);
  pthread_exit(NULL);

}


int
neblo_join
( reqst_t *reqst )
{

  ULL                   a,b;
  neblo_addr            foreign_id;
  neblo_addr            foreign_id_src;
  struct in_addr        foreign_addr;
  unsigned long         faddr;
  u_int16_t             foreign_port;
  unsigned int          fport;
  u_int32_t             foreign_join_handle_val;
  char                 *foreign_keymod;
  char                 *foreign_keyexp;
  unsigned char        *id_k;
  neblo_addr            id_chk;
  int                   rc = 0;
  peer_t               *hop;
  unsigned int          tmpunavail = 0;
  char                 *buf;
  pthread_t             th_id;


  _do_(_malloc_(foreign_keymod,KEYBUFLEN+SLACK),
       !foreign_keymod, "neblo_join(): foreign_keymod malloc");
  _do_(_malloc_(foreign_keyexp,BUFLEN+SLACK),
       !foreign_keyexp, "neblo_join(): foreign_keyexp malloc");

  sscanf(reqst->buf,"%*s %016llx %016llx %lu %u %u %50000s %100s",
         &a,&b,&faddr,&fport,&foreign_join_handle_val,
         foreign_keymod,foreign_keyexp);

  foreign_id = a;
  foreign_id_src = b;
  foreign_addr.s_addr = faddr;

  /*  SECURITY  */

  foreign_port = (u_int16_t)fport;
  if ((unsigned int)foreign_port != fport) {

#ifdef DEBUG_J
    S(fprintf(stderr,"[%lu]\tneblo_join(): bad foreign port. Abort...\n",
              reqst->th_id));
#endif

    goto abort;
  }

  /* SECURITY II:  check foreign id against its source */

  _do_(_malloc_(id_k,RIPEMD160_DIGEST_LENGTH+SLACK), !id_k,
       "neblo_join(): id_k malloc");
  RIPEMD160((unsigned char *)&foreign_id_src,sizeof(foreign_id_src),id_k);
  xor_fold((int8_t *)&id_chk,sizeof(id_chk),
           (int8_t *)id_k,RIPEMD160_DIGEST_LENGTH);
  _free_(id_k);
  if (id_chk != foreign_id) {

#ifdef DEBUG_J
    S(fprintf(stderr,"[%lu]\tneblo_join(): bad foreign id. Abort...\n",
              reqst->th_id));
#endif

    goto abort;
  }

  /* SECURITY III:  ring must be already enabled, refuse otherwise */

  if (!ring->enabled) {

#ifdef DEBUG_J
    S(fprintf(stderr,"[%lu]\tneblo_join(): ring not enabled yet. Abort...\n",
              reqst->th_id));
#endif

    goto abort;
  }

  /* from now on, mutual exclusion w.r.t. other joins */
  pthread_mutex_lock(&ring->join_lock);

  for (;;) {

    /*
     *  Find a LIVING peer P whose id. has minimal distance from
     *  the foreign_id in the ring ordering.
     *  Local peer is included in the search.
     */

    if ((hop = next_hop(foreign_id)) == NULL) {

#ifdef DEBUG_J
      S(fprintf(stderr,"[%lu]\tclosest is myself\n",reqst->th_id));
#endif

      /*  Local peer is the right one.  Must welcome foreign.  */

      if (tmpunavail)
        forall_peers(f,f->tmpunavail = 0);

      /*
       * Manage the newcomer by a separate thread; on exit, the thread itself
       * will deallocate "buf" and deassert the "join_lock".
       */

      _do_(_malloc_(buf,KEYBUFLEN+SLACK),!buf,"neblo_join(): buf malloc");
      memcpy(buf,reqst->buf,KEYBUFLEN);
      _do_(rc = pthread_create(&th_id,NULL,welcome_foreign,buf),rc,
           "neblo_join(): welcome_foreign() pthread_create");

#ifdef DEBUG_J
      S(fprintf(stderr,"[%lu]\twelcome_foreign() thread launched\n",
                reqst->th_id));
#endif

      _free_(foreign_keymod); _free_(foreign_keyexp);
      return 0;

    }

    if (hop == (peer_t *)(-1)) {

#ifdef DEBUG_J
      S(fprintf(stderr,"[%lu]\tneblo_join(): couldn't route. Abort...\n",
                reqst->th_id));
#endif

      rc = -1;
      break;
    }

#ifdef DEBUG_J
    S(fprintf(stderr,"[%lu]\tnext hop is %u.%u.%u.%u:%u (status %u)\n",
              reqst->th_id,
              ((unsigned char *)&hop->addr.s_addr)[0],
              ((unsigned char *)&hop->addr.s_addr)[1],
              ((unsigned char *)&hop->addr.s_addr)[2],
              ((unsigned char *)&hop->addr.s_addr)[3],
              hop->port,hop->status));
#endif

    sprintf(reqst->buf,"join %016llx %016llx %lu %u %u %s %s",
            (ULL)foreign_id,(ULL)foreign_id_src,
            (unsigned long)foreign_addr.s_addr,(unsigned int)foreign_port,
            foreign_join_handle_val,
            foreign_keymod,foreign_keyexp);

    rc = tell_peer(hop,is_successor(hop) ? 1 : 0,reqst->buf,
                   strlen(reqst->buf)+1,1,reqst->th_id);

    if (rc < 0) {
      /*
       *  Peer at next hop had troubles, might have been flagged as
       *  "tmpunavail"
       */
      tmpunavail = 1;
      continue;
    }

#ifdef DEBUG_J
    S(fprintf(stderr,"[%lu]\t(%u.%u.%u.%u:%u) << %s\n",
              reqst->th_id,
              ((unsigned char *)&hop->addr.s_addr)[0],
              ((unsigned char *)&hop->addr.s_addr)[1],
              ((unsigned char *)&hop->addr.s_addr)[2],
              ((unsigned char *)&hop->addr.s_addr)[3],
              hop->port,reqst->buf));
#endif

    break;

  }

  if (tmpunavail)
    forall_peers(f,f->tmpunavail = 0);

  /* resume accepting new join requests */
  pthread_mutex_unlock(&ring->join_lock);

  _free_(foreign_keymod); _free_(foreign_keyexp);
  return rc;

abort:

  _free_(foreign_keymod); _free_(foreign_keyexp);
  return -1;

}
