/*********************************************************************/
/* refresh.c                                                         */
/* function neblo_handle_refresh_msgs() and other related stuff      */
/* see COPYING in parent directory                                   */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>



/*
 * Send a msg of kind "refresh" for the addr interval "lo_addr" -- "up_addr",
 * to the peers mapping the addr interval "lo_key" -- "up_key", asking to send
 * their answers to the peer who maps address "req_addr"
 * (usually the originator).
 * Do not expect any answer.
 */
static int
send_refresh_request
(
  neblo_addr lo_key, neblo_addr up_key,
  neblo_addr lo_addr, neblo_addr up_addr,
  neblo_addr req_addr, pthread_t th_id
)
{

  unsigned char        *buf;
  int                   rc;
  peer_t               *hop;
  reqst_t              *req;
  unsigned int          tmpunavail = 0;


  _do_(_calloc_(buf,HEADERSIZE_F+SLACK),!buf,
       "send_refresh_request(): buf calloc");

  rc = 0;

  for (;;) {

    hop = next_hop(lo_key);

    if (hop == NULL) {
      /*
       * "lo_key" is mapped on the local peer.  This means the addr interval
       * lo_key -- up_key is at least partly mapped on the local peer.
       * It is useless to ask for a refresh of local data, so skip the local
       * part of addr interval and manage the residual non-local interval
       * (if any).
       */
      if (_belong_(up_key,ring->my_id,ring->up_bound))
        /*  All the interval is local -- nothing to do  */
        break;
      /*  Residual, non-local interval is ring->up_bound+1 -- up_key */
      lo_addr += ring->up_bound - lo_key + 1;
      lo_key = ring->up_bound + 1;

#ifdef DEBUG_RW
      S(fprintf(stderr,
                "[%lu]send_refresh_request(): using addr %016llx -- %016llx, keys %016llx -- %016llx\n",
                th_id,lo_addr,up_addr,lo_key,up_key));
#endif

      continue;
    }

    if (hop == (peer_t *)(-1)) {
      /*
       * "lo_key" is not mapped here, and this peer doesn't know where to
       * forward the request.
       */

      /*
       * insert the requested operation in a queue, to be tried again later
       */
      _do_(_calloc_(req,sizeof(reqst_t)),!req,
           "send_refresh_request(): req calloc");
      _do_(_calloc_(req->buf,KEYBUFLEN+SLACK),!req->buf,
           "send_refresh_request(): req->buf calloc");
      sprintf(req->buf,"refr %016llx %016llx %016llx %016llx %016llx ",
              (ULL)lo_key,(ULL)up_key,(ULL)lo_addr,(ULL)up_addr,(ULL)req_addr);
      req->th_id = 666; /* dummy th_id */
      insert_retry_queue(REFR,req,th_id);
      rc = -1;
      break;
    }

    /*
     * "lo_addr" is mapped remotely. this means the addr interval
     * lo_addr -- up_addr is not even partly mapped on the local peer.
     * route the refresh req through peer "hop"
     */

    sprintf(buf,"refr %016llx %016llx %016llx %016llx %016llx ",
            (ULL)lo_key,(ULL)up_key,(ULL)lo_addr,(ULL)up_addr,(ULL)req_addr);
    rc = tell_peer(hop,is_successor(hop) ? 1 : 0,buf,HEADERSIZE_F,1,th_id);
    if (rc >= 0) {

#ifdef DEBUG_RW
      S(fprintf(stderr,"[%lu]\t<< %s\n",th_id,buf));
#endif

      break;
    }

    /*
     * something wrong while talking to next peer,
     * peer migh have been flagged as "tmpunavail"
     */
    tmpunavail = 1;
    /* choose another peer */
    continue;

  }

  if (tmpunavail)
    forall_peers(f,f->tmpunavail = 0);

  _free_(buf);
  return rc;

}



/*
 *  Issue "refresh" msgs, to rebuild the content of a given addr intvl.
 *  The intvl contains primary data associated to the addresses, as well as
 *  backup data associated to other addr intvls (related to the given intvl
 *  through a simple arithmetics rule).
 */
void
ask_refresh_addrs
( neblo_addr lo_bound, neblo_addr up_bound, unsigned int cp, pthread_t th_id )
{

  neblo_addr          key_lo;
  neblo_addr          key_up;
  neblo_addr          addr_lo;
  neblo_addr          addr_up;
  unsigned int        i;
  unsigned int        j;


  addr_lo = lo_bound;
  addr_up = up_bound;

  for (i = 0; i < cp; i++) {

    addr_lo -= ((neblo_addr)1)<<(8*sizeof(neblo_addr)-i);
    addr_up -= ((neblo_addr)1)<<(8*sizeof(neblo_addr)-i);
    /*  when i=0, addr_lo and addr_up are same as lo_bound and up_bound.  */

    key_lo = addr_lo;
    key_up = addr_up;

    for (j=0; j < cp; j++) {

      key_lo += ((neblo_addr)1)<<(8*sizeof(neblo_addr)-j);
      key_up += ((neblo_addr)1)<<(8*sizeof(neblo_addr)-j);
      /*  when j=0, key_lo and key_up are same as addr_lo and addr_up.  */

      /*
       *  Send a "refr" req for addr intvl addr_lo -- addr_up
       *  to peers who map the addr intvl key_lo -- key_up.
       *  Avoid sending a "refr" req to self.
       */
      if (j == i)
        continue;
      send_refresh_request(key_lo,key_up,addr_lo,addr_up,lo_bound,th_id);
    }

  }

  return;

}



/*
 *  Ask the application layer to re-publish (towards a given addr) the local
 *  content associated to a given addr intvl.
 *  The addr intvl contains primary data associated to the addresses, as well
 *  as backup data associated to other addr intvls (related to the given intvl
 *  through a simple arithmetics rule).
 */
int
republish_addrs
( neblo_addr lo_bound, neblo_addr up_bound, neblo_addr addr, unsigned int cp )
{

  neblo_addr          addr_lo;
  neblo_addr          addr_up;
  neblo_addr          past_addr_lo;
  neblo_addr          past_addr_up;
  neblo_addr          a_l;
  neblo_addr          a_u;
  unsigned int        i;
  int                 rc = 0;
  int                 r;


  r = invoke_refresh_handler(lo_bound,up_bound,addr);
  if (r < 0)
    rc = r;

  past_addr_lo = lo_bound;
  past_addr_up = up_bound;

  for (i = 1; i < cp; i++) {

    addr_lo = past_addr_lo - (((neblo_addr)1)<<(8*sizeof(neblo_addr)-i));
    addr_up = past_addr_up - (((neblo_addr)1)<<(8*sizeof(neblo_addr)-i));

    /*  avoid overlapping.  a bit involved...  */
    a_l = addr_lo;
    a_u = addr_up;
    if (_belong_(a_l,past_addr_lo,past_addr_up))
      a_l = past_addr_up;
    if (_belong_(a_u,past_addr_lo,past_addr_up))
      a_u = past_addr_lo;
    if (_belong_(a_l,lo_bound,up_bound) && _belong_(a_u,lo_bound,up_bound))
      break;
    if (_belong_(a_l,lo_bound,up_bound))
      a_l = up_bound;
    if (_belong_(a_u,lo_bound,up_bound))
      a_l = lo_bound;

    r = invoke_refresh_handler(a_l,a_u,addr);
    if (r < 0)
      rc = r;

    past_addr_lo = addr_lo;
    past_addr_up = addr_up;

  }

  return rc;

}



/*
 *  Ask the application layer to delete the local content associated to a
 *  given addr intvl.
 *  The addr intvl contains primary data associated to the addresses, as well
 *  as backup data associated to other addr intvls (related to the given intvl
 *  through a simple arithmetics rule).
 */

typedef struct cl_t {
  neblo_addr   l;
  neblo_addr   u;
  int          dontclean;
  struct cl_t *next;
} clean_interval_t;

int
clean_addrs
( neblo_addr lo_bound, neblo_addr up_bound, unsigned int cp )
{

  clean_interval_t   *clean_list;
  clean_interval_t   *c;
  clean_interval_t   *d;
  unsigned int        i;
  unsigned int        j;
  neblo_addr          addr_lo;
  neblo_addr          addr_up;
  neblo_addr          u;
  int                 rc = 0;
  int                 r;


  /*  Build a list of cleanable intervals of addresses.  A bit involved...  */

  _do_(_malloc_(c,sizeof(clean_interval_t)),!c,"clean_addrs(): c malloc");
  clean_list = c;
  c->l = lo_bound;
  c->u = up_bound;
  c->dontclean = 0;
  c->next = NULL;

  for (i = 0; i < cp; i++) {

    if (i > 0) {
      lo_bound += ((neblo_addr)1)<<(8*sizeof(neblo_addr)-i);
      up_bound += ((neblo_addr)1)<<(8*sizeof(neblo_addr)-i);
      _do_(_malloc_(d,sizeof(clean_interval_t)),!d,"clean_addrs(): d malloc");
      d->l = lo_bound;
      d->u = up_bound;
      d->dontclean = 0;
      d->next = NULL;
      c->next = d;
      c = d;
    }

    /*  It is useless to turn around the ring  */
    if (_belong_(ring->my_id,c->l,c->u)) {
      c->u = ring->my_id-1;
    }
    if (_belong_(c->l,ring->my_id,ring->up_bound)) {
      c->dontclean = 1;
      break;
    }

    /*  Do not clean over regions of possible backup locations  */

    addr_lo = ring->my_id;
    addr_up = ring->up_bound;
    for (j=1; j < (i+1) % cp; j++) {
      addr_lo += ((neblo_addr)1)<<(8*sizeof(neblo_addr)-j);
      addr_up += ((neblo_addr)1)<<(8*sizeof(neblo_addr)-j);
    }
    for (; j != i; j = (j+1) % cp) {
      if (j) {
        addr_lo += ((neblo_addr)1)<<(8*sizeof(neblo_addr)-j);
        addr_up += ((neblo_addr)1)<<(8*sizeof(neblo_addr)-j);
      } else {
        addr_lo = ring->my_id;
        addr_up = ring->up_bound;
      }
      if (_belong_(c->l,addr_lo,addr_up)) {
        if (_belong_(c->u,addr_lo,addr_up)) {
          /* ignore clean intvl because fully contained into a backup intvl */
          c->dontclean = 1;
          goto next;
        }
        /* shrink clean intvl because partly contained into a backup intvl */
        c->l = addr_up+1;
      }
      if (_belong_(addr_lo,c->l,c->u)) {
        /* split clean intvl because of overlapping with backup intvl */
        u = c->u;
        c->u = addr_lo-1;
        if (_belong_(addr_up+1,addr_lo,u)) {
          _do_(_malloc_(d,sizeof(clean_interval_t)),!d,
               "clean_addrs(): d malloc");
          d->l = addr_up+1;
          d->u = u;
          d->dontclean = 0;
          d->next = NULL;
          c->next = d;
          c = d;
        }
      }
    }

next:

    /* this is here to inhibit a silly complaint of gcc */
    continue;

  }

  /*  Purge the clean list from duplicated entries  */

  for (c=clean_list; c!=NULL; c = c->next) {
    if (c->dontclean) continue;
    for (d=clean_list; d!=c; d = d->next) {
      if (d->dontclean) continue;
      if (c->l == d->l && c->u == d->u) {
        c->dontclean = 1;
        break;
      }
    }
  }

  for (c=clean_list; c!=NULL; c = c->next) {
    if (c->dontclean) continue;
    r = invoke_clean_handler(c->l,c->u);
    if (r < 0)
      rc = r;
  }

  return rc;

}



/*
 * this peer is asked to take a "refresh" request for a given addr interval,
 * and deliver to application if this peer is the final destination, or
 * forward on otherwise.
 */
int
neblo_handle_refresh_msgs
( reqst_t *reqst, unsigned int retry )
{

  ULL                   a,b,c,d,e;
  neblo_addr            lo_key;
  neblo_addr            up_key;
  neblo_addr            lo_addr;
  neblo_addr            up_addr;
  neblo_addr            req_addr;
  int                   rc;
  peer_t               *hop;
  neblo_addr            key;
  neblo_addr            addr;
  unsigned int          tmpunavail = 0;


  sscanf(reqst->buf,"%*s %016llx %016llx %016llx %016llx %016llx",
         &a,&b,&c,&d,&e);
  lo_key = a; up_key = b; lo_addr = c; up_addr = d; req_addr = e;

#ifdef DEBUG_RW
  if (retry != FIRST_ATTMPT)
    S(fprintf(stderr,
              "[%lu]\tRETRY refr: addrs %016llx -- %016llx, keys %016llx -- %016llx\n",
              reqst->th_id,lo_addr,up_addr,lo_key,up_key));
#endif

  /*
   * SECURITY: requests of this kind may only be accepted
   * if coming from a friend.
   * however, if this is a retry, the request might be from the
   * local peer; in this case it must be accepted
   */
  if (!is_friend(reqst->csock) && retry == FIRST_ATTMPT) {

#ifdef DEBUG_RW
    S(fprintf(stderr,"[%lu]\tREFUSED refresh: not a friend\n",reqst->th_id));
#endif

    return -1;
  }

  rc = 0;

  for (;;) {

    hop = next_hop(lo_key);

    if (hop == NULL) {
      /*
       * "lo_key" is mapped locally. this means the addr interval
       * lo_key -- up_key is at least partly mapped on the local peer.
       */
      /* determine the locally mapped subinterval */
      if (_belong_(up_key,ring->my_id,ring->up_bound)) {
        key = up_key;
        addr = up_addr;
      } else {
        key = ring->up_bound;
        addr = up_addr - (up_key - ring->up_bound);
      }
      /*
       *  Deliver refresh request to the application layer, limited to
       *  the local part of addr interval (possibly the entire interval),
       *  but only if the originator was not the local peer itself.
       */
      if (next_hop(req_addr) != NULL)
        rc = invoke_refresh_handler(lo_addr,addr,req_addr);
      /*  Manage the residual non-local interval, if not empty.  */
      if (addr != up_addr) {
        lo_key = key + 1;
        lo_addr = addr + 1;
        sprintf(reqst->buf,"refr %016llx %016llx %016llx %016llx %016llx",
               (ULL)lo_key,(ULL)up_key,(ULL)lo_addr,(ULL)up_addr,
               (ULL)req_addr);
        continue;
      }
      /* done. return */
      break;
    }

    if (hop == (peer_t *)(-1)) {
      /*
       * blk "lo_addr" is not here, and this peer doesn't know where to forward
       * the request at the moment.
       * insert the requested operation in a queue, to be tried again later,
       * but only if this is not a retry in itself (otherwise we could be
       * inserting the operation multiple times)
       */
      if (retry == FIRST_ATTMPT) {
        /*
         * we must do a copy. the original reqst buffer will be put back
         * in the service queue, the copy is to be put in the retry queue
         */
        reqst_t *r;

        _do_(_malloc_(r,sizeof(reqst_t)),!r,
             "neblo_handle_refresh_msgs(): r malloc");
        memcpy(r,reqst,sizeof(reqst_t));
        _do_(_calloc_(r->buf,KEYBUFLEN+SLACK),!r->buf,
             "neblo_handle_refresh_msgs(): r->buf calloc");
        memcpy(r->buf,reqst->buf,HEADERSIZE_F);
        insert_retry_queue(REFR,r,reqst->th_id);
      } else
        if (retry == LAST_CHANCE) {
          /* no more chances.  Msg is dropped :-( */
        }
      rc = -1;
      break;
    }

    /*
     * "lo_key" is mapped remotely. this means the addr interval
     * lo_key -- up_key is not even partly mapped on the local peer.
     * route the refresh req through peer "hop"
     */

    rc = tell_peer(hop,is_successor(hop) ? 1 : 0,reqst->buf,HEADERSIZE_F,1,
                   reqst->th_id);
    if (rc >= 0) {

#ifdef DEBUG_RW
      S(fprintf(stderr,"[%lu]\t<< %s\n",reqst->th_id,reqst->buf));
#endif

      /* done. return */
      break;
    }

    /*
     * something wrong while talking to next peer, might have been flagged as
     * "tmpunavail"
     */

#ifdef DEBUG_RW
    S(fprintf(stderr,
              "[%lu]\tforward refr to peer %u.%u.%u.%u:%u (%u) returned %d\n",
              reqst->th_id,
              ((unsigned char *)&hop->addr.s_addr)[0],
              ((unsigned char *)&hop->addr.s_addr)[1],
              ((unsigned char *)&hop->addr.s_addr)[2],
              ((unsigned char *)&hop->addr.s_addr)[3],
              hop->port,hop->status,
              rc));
#endif

    tmpunavail = 1;
    /* choose another peer */
    continue;

  }

  /*
   * the tell_peer() routine above might have flagged some peers as
   * "tmpunavail"...
   */
  if (tmpunavail)
    forall_peers(f,f->tmpunavail = 0);

  return rc;

}
